"""
- test the databricks api token
- zip the test archive
- walk the test folders to determine the parallelization
- constuct the test job json
    - tag the test job for easy retrieval
- submit the test job
- return the ID - also to file
"""
import argparse
import copy
import datetime
import inspect
import json
import re
import shutil
import subprocess
import tempfile
import time
import uuid
from pathlib import Path
from typing import Dict, List, Union
from typing.io import IO

from . import test_main
from .dbcli import db_check, dbfscall, dbjcall
from .dbfs import DbfsLocation


def setup_submit_parser(subparsers):
    """
    Adds a subparser for the command 'submit'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """

    parser: argparse.ArgumentParser = subparsers.add_parser(
        "submit", description="Run Test Cases on databricks cluster."
    )
    parser.set_defaults(func=submit_main)

    parser.add_argument(
        "--wheels",
        type=str,
        required=False,
        help="The glob paths of all wheels under test.",
        default="dist/*.whl",
    )

    parser.add_argument(
        "--tests",
        type=str,
        required=True,
        help="Location of the tests folder. Will be sendt to databricks as a whole.",
    )

    task = parser.add_mutually_exclusive_group(required=True)
    task.add_argument(
        "--task",
        help="Single Test file or folder to execute.",
    )
    task.add_argument(
        "--tasks-from",
        help="path in test archive where each subfolder becomes a task.",
    )

    # cluster argument pair
    cluster = parser.add_mutually_exclusive_group(required=True)
    cluster.add_argument(
        "--cluster",
        type=str,
        help="JSON document describing the cluster setup.",
    )
    cluster.add_argument(
        "--cluster-file",
        type=argparse.FileType("r"),
        help="File with JSON document describing the cluster setup.",
    )

    # spark libraries argument pair
    sparklibs = parser.add_mutually_exclusive_group(required=False)
    sparklibs.add_argument(
        "--sparklibs",
        help="JSON document describing the spark dependencies.",
    )
    sparklibs.add_argument(
        "--sparklibs-file",
        type=argparse.FileType("r"),
        help="File with JSON document describing the spark dependencies.",
    )

    # python dependencies file
    pydep = parser.add_mutually_exclusive_group(required=False)
    pydep.add_argument(
        "--requirement",
        action="append",
        help="a python dependency, specified like for pip",
        default=[],
    )
    pydep.add_argument(
        "--requirements-file",
        type=argparse.FileType("r"),
        help="File with python dependencies, specified like for pip",
    )

    parser.add_argument(
        "--main-script",
        type=argparse.FileType("r"),
        help="Your own test_main.py script file, to add custom functionality.",
    )

    parser.add_argument(
        "--pytest-args", help="Additional arguments to pass to pytest in each test job."
    )

    parser.add_argument(
        "--out-json",
        type=argparse.FileType("w"),
        help="File to store the RunID for future queries.",
    )

    return


def collect_arguments(args):
    """
    Post process the parsed arguments of the 'submit' command argument parser.
    :param args: parsed arguments of the 'submit' command argument parser
    :return:
    """

    # pre-process 'cluster'
    if args.cluster_file:
        args.cluster = args.cluster_file.read()
    args.cluster = json.loads(args.cluster)

    # pre-process 'sparklibs'
    if args.sparklibs_file:
        args.sparklibs = args.sparklibs_file.read()
    if args.sparklibs:
        args.sparklibs = json.loads(args.sparklibs)

    # pre-process 'requirement'
    if args.requirements_file:
        args.requirement = [
            line.strip()
            for line in args.requirements_file.read().splitlines()
            if line.strip() and not line.strip().startswith("#")
        ]

    args.pytest_args = (args.pytest_args or "").split()
    if args.tasks_from:
        args.parallelize = True
        args.task = args.tasks_from
    else:
        args.parallelize = False

    return args


def submit_main(args):
    """the main function of the cli command 'submit'. Not to be used directly."""
    db_check()

    args = collect_arguments(args)

    submit(
        test_path=args.tests,
        task=args.task,
        parallelize=args.parallelize,
        cluster=args.cluster,
        wheels=args.wheels,
        requirement=args.requirement,
        sparklibs=args.sparklibs,
        out_json=args.out_json,
        main_script=args.main_script,
        pytest_args=args.pytest_args,
    )


class DbTestFolder:
    """Context manager that creates a unique test folder on dbfs."""

    def __init__(self):
        self._test_path_base = DbfsLocation(
            "/".join(
                [
                    "test",
                    datetime.datetime.now(datetime.timezone.utc).strftime(
                        "%Y%m%d-%H%M%S"
                    ),
                    uuid.uuid4().hex,
                ]
            )
        )

    def __enter__(self):
        print(f"Making dbfs test folder {self._test_path_base.remote}")
        dbfscall(f"mkdirs {self._test_path_base.remote}")
        return self._test_path_base

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def verify_and_resolve_task(test_path: str, task: Union[str, Path]):
    test_archive = Path(test_path).resolve().absolute()

    task_path = Path(task)
    if not task_path.is_absolute():
        task_path = (test_archive.parent / task_path).resolve().absolute()

    task_path = task_path.resolve().absolute()

    if not (test_archive == task_path or test_archive in task_path.parents):
        raise AssertionError(f"task {task} is not contained in {test_path}")

    if not task_path.exists():
        raise AssertionError(f"task {task_path} does not exist")

    if not (test_archive.parent / task).exists():
        raise AssertionError(f"The task {task} was not found in the test location.")

    return task_path.relative_to(test_archive.parent).as_posix()


def discover_job_tasks(test_path: str, folder: str):
    """If folder is given, create parallel tasks from this level.
    Otherwise, simply return the top folder as the single task.
    Returns a list of strings with the subfolders to process.
    """

    test_archive_parent = Path(test_path).resolve().absolute().parent

    subfolders = [
        verify_and_resolve_task(test_path, x)
        for x in (test_archive_parent / folder).iterdir()
        if x.is_dir()
    ]
    return subfolders


class PoolBoy:
    """Hold a list of available instance pools and replace the by-name reference with an id if possible."""

    MARKER = "instance-pool://"

    def __init__(self):
        self._lookup = self.get_instance_pools()

    def lookup(self, pool_id: str) -> str:
        if pool_id.startswith(self.MARKER):
            pool_name = pool_id[len(self.MARKER) :]
            return self._lookup[pool_name]
            # if this throws a KeyError, we are right to abort execution since the input is invalid.
        else:
            # No Marker = no lookup
            return pool_id

    def get_instance_pools(self) -> Dict[str, str]:
        pool_data = dbjcall("instance-pools list --output JSON")
        pool_lookup = {
            pool["instance_pool_name"]: pool["instance_pool_id"]
            for pool in pool_data["instance_pools"]
        }
        return pool_lookup


def submit(
    test_path: str,
    task: str,
    cluster: dict,
    wheels: str,
    parallelize: bool,
    requirement: List[str] = None,
    sparklibs: List[dict] = None,
    out_json: IO[str] = None,
    main_script: IO[str] = None,
    pytest_args: List[str] = None,
):
    if requirement is None:
        requirement = []
    if sparklibs is None:
        sparklibs = []
    if pytest_args is None:
        pytest_args = []

    # check the structure of the cluster object
    if not isinstance(cluster, dict):
        raise AssertionError("invalid cluster specification")

    # check the structure of the sparklibs object
    if not isinstance(sparklibs, list):
        raise AssertionError("invalid sparklibs specification")

    for py_requirement in requirement:
        sparklibs.append({"pypi": {"package": py_requirement}})

    pools = PoolBoy()

    with DbTestFolder() as test_folder:
        for wheel in discover_and_push_wheels(wheels, test_folder):
            sparklibs.append({"whl": wheel})

        archive_local = archive_and_push(test_path, test_folder)
        main_file = push_main_file(test_folder, main_script)

        print(f"copied everything to {test_folder.remote}")

        print("Wait 1 minute for uploading test folder...")
        time.sleep(secs=60)
        print("Waited 1 minute successfully!")

        # construct the workflow object
        workflow = dict(run_name="Testing Run", format="MULTI_TASK", tasks=[])

        if parallelize:
            # subtasks will be ['tests/cluster/job1', 'tests/cluster/job2'] or similar
            tasks = discover_job_tasks(test_path, task)
        else:
            tasks = [verify_and_resolve_task(test_path, task)]

        for task in tasks:
            task_sub = re.sub(r"[^a-zA-Z0-9_-]", "_", task)

            # prepare cluster
            task_cluster = copy.deepcopy(cluster)
            task_cluster["cluster_log_conf"] = {
                "dbfs": {"destination": f"{test_folder.remote}/{task_sub}"}
            }
            if "instance_pool_id" in task_cluster:
                task_cluster["instance_pool_id"] = pools.lookup(
                    task_cluster["instance_pool_id"]
                )

            workflow["tasks"].append(
                dict(
                    task_key=task_sub,
                    libraries=sparklibs,
                    spark_python_task=dict(
                        python_file=main_file.remote,
                        parameters=[
                            # running in the spark python interpreter, the python __file__ variable does not
                            # work. Hence, we need to tell the script where the test area is.
                            f"--basedir={test_folder.local}/{task_sub}",
                            # all tests will be unpacked from here
                            f"--archive={archive_local}",
                            # we can actually run any part of our test suite, but some files need the full repo.
                            # Only run tests from this folder.
                            f"--folder={task}",
                            # additional arguments to pass to pytest
                            f"--pytestargs={json.dumps(pytest_args)}",
                        ],
                    ),
                    new_cluster=task_cluster,
                )
            )

    print("Submitting test...")
    with tempfile.TemporaryDirectory() as tmp:
        jobfile = f"{tmp}/job.json"
        with open(jobfile, "w") as f:
            json.dump(workflow, f)

        try:
            res = dbjcall(f"runs submit --json-file {jobfile}")
        except subprocess.CalledProcessError:
            print("Json contents:")
            print(json.dumps(workflow, indent=4))
            raise

        try:
            run_id = res["run_id"]
        except KeyError:
            print(res)
            raise

    # now we have the run_id
    print(f"Started run with ID {run_id}")
    details = dbjcall(f"runs get --run-id {run_id}")
    print(f"Follow job details at {details['run_page_url']}")

    if out_json:
        json.dump({"run_id": run_id}, out_json)


def discover_and_push_wheels(globpath: str, test_folder: DbfsLocation) -> List[str]:
    result = []
    for item in Path().glob(globpath):
        remote_path = f"{test_folder.remote}/{item.parts[-1]}"
        print(f"pushing {item} to test folder")
        dbfscall(f"cp {item} {remote_path}")
        result.append(remote_path)

    return result


def archive_and_push(test_path: str, test_folder: DbfsLocation):
    with tempfile.TemporaryDirectory() as tmp:
        test_path = Path(test_path)
        print(f"now archiving {test_path}")
        archive_path = shutil.make_archive(
            str(Path(tmp) / "tests"),
            "zip",
            test_path / "..",
            base_dir=test_path.parts[-1],
        )
        print("now pushing test archive to test folder")

        dbfscall(f"cp {archive_path} {test_folder.remote}/tests.zip")

    return f"{test_folder.local}/tests.zip"


def push_main_file(
    test_folder: DbfsLocation, main_script: IO[str] = None
) -> DbfsLocation:
    print("now pushing test main file")
    main_file = test_folder / "main.py"
    with tempfile.TemporaryDirectory() as tmp:
        with open(Path(tmp) / "main.py", "w") as f:
            if main_script:
                f.write(main_script.read())
            else:
                print("Using default main script test_main.py")
                f.write(inspect.getsource(test_main))

        dbfscall(f"cp {tmp}/main.py {main_file.remote}")

    return main_file



from typing import Any, List, Union

from pyspark.sql import DataFrame
from spetlr.tables.TableHandle import TableHandle


class TestHandle(TableHandle):
    __test__ = False  # solves PytestCollectionWarning

    def __init__(self, provides: DataFrame = None):
        self.provides = provides
        self.overwritten = None
        self.appended = None
        self.truncated = False
        self.dropped = False
        self.dropped_and_deleted = False
        self.upserted = None
        self.upserted_join_cols = None
        self.comparison_col = None
        self.comparison_limit = None
        self.comparison_operator = None

    def read(self) -> DataFrame:
        if self.provides is not None:
            return self.provides
        else:
            raise AssertionError("TableHandle not readable.")

    def overwrite(self, df: DataFrame) -> None:
        self.overwritten = df

    def append(self, df: DataFrame) -> None:
        self.appended = df

    def truncate(self) -> None:
        self.truncated = True

    def drop(self) -> None:
        self.dropped = True

    def drop_and_delete(self) -> None:
        self.dropped_and_deleted = True

    def upsert(self, df: DataFrame, join_cols: List[str]) -> Union[DataFrame, None]:
        self.upserted = df
        self.upserted_join_cols = join_cols

    def delete_data(
        self, comparison_col: str, comparison_limit: Any, comparison_operator: str
    ) -> None:
        self.comparison_col = comparison_col
        self.comparison_limit = comparison_limit
        self.comparison_operator = comparison_operator



import json

import pyspark.sql.functions as f
import pyspark.sql.types as t
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import StructType
from spetlr.spark import Spark


class ExtractEncodedBody:
    def __init__(
        self,
        json_field: str = "Body",
        new_column_w_extracted_body: str = None,
        data_limit: int = None,
    ):
        self.json_field = json_field
        self.new_field_w_extracted_body = new_column_w_extracted_body or json_field
        self._data_limit = data_limit
        self._fields_same_name = self.json_field == self.new_field_w_extracted_body

    def extract_schema(self, df: DataFrame) -> StructType:
        assert self.json_field in df.columns, "The body column is not in the dataframe"

        json_data = self._extract_using_rdd(df)

        return Spark.get().read.json(json_data).schema

    def extract_schema_as_json(self, df: DataFrame, pretty_json: bool = False) -> str:
        schema = self.extract_schema(df)

        # load the json as a data frame. Schema is inferred from data.
        json_schema_raw = schema.jsonValue()

        return json.dumps(json_schema_raw, indent=4 if pretty_json else None)

    def _extract_using_rdd(self, df: DataFrame):
        # extract json data rows using name of json_field
        json_data = df.withColumn(
            self.json_field, f.col(self.json_field).cast(t.StringType())
        )

        # Limiting the data used for extracting schema
        if self._data_limit:
            json_data = json_data.limit(self._data_limit)

        # extract json data rows using name of json_field
        json_data = json_data.rdd.map(lambda row: row.asDict()[self.json_field])

        return json_data

    def transform_df(
        self, df: DataFrame, keep_original_body: bool = False
    ) -> DataFrame:
        if self._fields_same_name and keep_original_body:
            raise ValueError(
                f"The field {self.json_field} is overwritten by the extracted json. Give extracted field new name to "
                f"keep original body. "
            )

        # Step 1: extract the schema
        schema = self.extract_schema(df)

        # Step 2: decode the body column
        df_transformed = df.withColumn(
            self.new_field_w_extracted_body,
            f.from_json(f.decode(self.json_field, "utf-8"), schema),
        )

        # Step 3: Drop the original column
        if (not keep_original_body) and (not self._fields_same_name):
            df_transformed = df_transformed.drop(self.json_field)

        return df_transformed



from html.parser import HTMLParser
from io import StringIO


class HTMLStripper(HTMLParser):
    """removes all HTML markup, leaving only normal letters and characters."""

    # Drawio nodes use HTML markup. I want this markup to not affect comparisons.
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def handle_data(self, d):
        # needed for the internal working of HTMLParser
        # not to be used directly
        self.text.write(d)
        return self

    def get_data(self) -> str:
        return self.text.getvalue()

    @staticmethod
    def strip(text: str) -> str:
        """Return the input minus any markup."""
        o = HTMLStripper()
        o.feed(text)
        return o.get_data()



import argparse
import configparser
import sys
from typing import IO

import requests
from packaging.version import Version, parse


def main():
    """Main CLI function, not to be used directly."""
    parser = argparse.ArgumentParser(
        description="Automatically set the version for upload to pypi"
    )
    parser.add_argument(
        "-t",
        dest="prerelease",
        action="store_true",
        help="prepare pre-release version for test.pypi",
    )
    parser.add_argument(
        "--name", help="Package name, if different from name in setup.cfg"
    )
    parser.add_argument(
        "--version-file",
        default="src/VERSION.txt",
        help="location of version to manipulate",
    )

    args = parser.parse_args()
    name = args.name
    if not name:
        config = configparser.ConfigParser()
        config.read("setup.cfg")
        name = config["metadata"]["name"]

    try:
        with open(args.version_file, "r+", encoding="utf-8") as v_file:
            mainpulate_version(
                package_name=name, version_file=v_file, pre_release=args.prerelease
            )

    except FileNotFoundError:
        print(f"Version file {args.version_file} not found.", file=sys.stderr)
        parser.print_usage()
        exit(-1)


def mainpulate_version(package_name: str, version_file: IO, pre_release: bool):
    """Query the indices, manipuate the version file and print the vesions to screen."""

    # get all relevant versions for manipulation
    test_v = get_remote_version("test.pypi.org", package_name)
    pypi_v = get_remote_version("pypi.org", package_name)
    local_v = get_local_version(version_file)

    next_version = _decide_next_version(pypi_v, test_v, local_v, pre_release)

    version_file.seek(0)
    version_file.truncate()
    version_file.write(str(next_version))
    print(str(next_version))


class InvalidVersionManipulation(Exception):
    pass


def _decide_next_version(
    pypi_v: Version, test_v: Version, local_v: Version, pre_release: bool
):
    """Returns the next version to be published in the local flow.
    See the documentation for how versions are manipulated.
    """

    # simple incrementation yields the naive next release
    next_incremental_release = parse(f"{pypi_v.major}.{pypi_v.minor}.{pypi_v.micro+1}")

    # if local_v is higher than a simple increment, that is what we use
    next_release = max(next_incremental_release, local_v)

    if not pre_release:
        return next_release

    # in case of pre-release, we need more work.
    if parse(test_v.base_version) < next_release:
        # the last test.pypi version relates to a previous release candidate
        # start a new serios of pre-releases
        pre_n = 1
    elif test_v.release == next_release.release and test_v < next_release:
        # we have previously released a pre-release version to test_pypi.
        # just use the next one
        if test_v.pre is not None:
            abrc, n = test_v.pre
            if abrc == "rc":
                # in case of a(lpha) or b(eta) release, still count the rc from 1
                pre_n = n + 1
    else:
        raise InvalidVersionManipulation(
            f"Test Pypi has {test_v}, "
            f"and next final release is {next_release}. "
            f"Releasing a release candidate for {next_release} "
            "would not make that release caindiate the highest on test.pypi. "
            "Please increment you package version to above the test.pypi version "
            "and try again."
        )

    return parse(f"{next_release.base_version}rc{pre_n}")


def get_remote_version(server: str, package: str):
    """Query the repository and return the Version (PEP-0440)
    that is the highest of any kind, not just final versions.
    In case of any issues, return Version("0"), which is lower
    than any other version and therefores usitable for the rest of the logic."""
    try:
        return max(
            parse(k)
            for k in requests.get(f"https://{server}/pypi/{package}/json")
            .json()["releases"]
            .keys()
        )
        # return parse(remote["info"]["version"])
    except:  # noqa: E722  # bare except is fine in this simple case
        return parse("0")


def get_local_version(v_file: IO):
    """read the version from the file. Only keep the base part of the version string."""
    v_file.seek(0)
    v = parse(v_file.read())
    v = parse(v.base_version)  # remove any suffices
    return v



# !/usr/bin/env python
import argparse
import configparser
import json
import re
import sys
import tempfile
import venv
from io import TextIOBase
from pathlib import Path
from subprocess import run
from typing import List, TypedDict


class LibType(TypedDict):
    name: str
    version: str


def main():
    parser = argparse.ArgumentParser(
        description="Update requirement versions in specified file."
    )
    parser.add_argument(
        "in_file",
        type=str,
        help="The requirements file to read.",
    )
    parser.add_argument(
        "-o, --out-file",
        type=str,
        default=None,
        dest="out_file",
        help="The output to file.",
    )
    parser.add_argument("--cfg", action="store_const", const=True, default=False)
    parser.add_argument(
        "--cfg-file",
        type=str,
        default="setup.cfg",
        dest="cfg_file",
        help="specify your configuration file if it differs from setup.cfg",
    )
    parser.add_argument(
        "--reject",
        type=str,
        default="pip|pywin32",
        help="regex to exclude. Default: pip|pywin32",
    )
    args = parser.parse_args()

    if not args.out_file and not args.cfg:
        just_print = True
        debug_out = sys.stderr
    else:
        just_print = False
        debug_out = sys.stdout

    # validate arguments:
    if args.out_file:
        with open(args.out_file, "w") as f:
            pass
        print(f"Output will be in {args.out_file}.")
    if args.cfg:
        print(f"Output will be in {args.cfg_file}.")

    with open(args.in_file) as f:
        freeze = freeze_req(f.read(), reject=args.reject, debug_out=debug_out)

    if args.out_file:
        with open(args.out_file, "w") as f:
            f.write("\n".join(f"{lib['name']}=={lib['version']}" for lib in freeze))

    if args.cfg:
        injection = "".join(f"\n{lib['name']}=={lib['version']}" for lib in freeze)
        config = configparser.ConfigParser()
        config.read(args.cfg_file)
        config["options"]["install_requires"] = injection
        with open(args.cfg_file, "w") as f:
            config.write(f)

    if just_print:
        print("\n".join(f"{lib['name']}=={lib['version']}" for lib in freeze))


def freeze_req(
    requirements: str, reject: str = "", debug_out: TextIOBase = sys.stdout
) -> List[LibType]:
    pat = re.compile(reject)

    with tempfile.TemporaryDirectory() as td:
        builder = venv.EnvBuilder(with_pip=True)
        builder.create(td)

        rf_path = str(Path(td) / "reqs.txt")
        with open(rf_path, "w") as f:
            f.write(requirements)

        python = str(Path(td) / "bin" / "python")
        if not Path(python).exists():
            python = str(Path(td) / "Scripts" / "python.exe")

        run(
            [python, "-m", "pip", "install", "--upgrade", "pip"],
            stdout=debug_out.buffer,
        )

        run([python, "-m", "pip", "install", "-r", rf_path], stdout=debug_out.buffer)

        freeze_json = run(
            [python, "-m", "pip", "list", "--format", "json"],
            capture_output=True,
            text=True,
        )
        freeze = json.loads(freeze_json.stdout)
        return [lib for lib in freeze if not pat.match(lib["name"])]



import re
from typing import List

from pyspark.sql import DataFrame


def validate_camelcased_cols(
    df: DataFrame, cols_to_check: List[str] = None, print_result: bool = False
):
    """
    The function takes a dataframe and validates if all columns or a given subset of columns are camelCased.

    :param df: A pyspark dataframe
    :param cols_to_check: Optional, a subset of the columns in the dataframe to check.
    :param print_result: If True, then the cols which are not camelCased are printed
    :return: True if all columns are camelcased
    """

    if cols_to_check is None:
        cols_to_check = df.columns
    else:
        if not all(col in df.columns for col in cols_to_check):
            raise ValueError("Some of the columns to check is not in the dataframe.")

    results_col = []
    check_passed = True
    for col in cols_to_check:
        is_camel_cased = bool(re.match("[a-z]+([A-Z][a-zA-Z]*)?", col))
        has_two_upper = bool(re.match(".*[A-Z][A-Z]", col))
        if not is_camel_cased or has_two_upper:
            check_passed = False
            results_col.append(col)

    # Todo : print results
    if print_result:
        print(
            f"The following columns are not camelCased: {results_col}"
            if any(results_col)
            else "All tested columns are camelCased!"
        )

    return check_passed
